/*
 * chaucer.ino
 *
 * Demonstrates how the urboot bootloader can be used to overwrite PROGMEM during runtime.
 * urbootPageWrite() is a wrapper that calls the pgm_write_page() bootloader code.
 *
 *
 * published under GNU General Public License, version 3 (GPL-3.0)
 * author Stefan Rueger
 *
 * v 1.2
 * first version 13.07.2016
 */


// #include <avr/pgmspace.h>
#include <config.h>
#include "urboot.h"
// #include "chaucer.h"
#include "../ur/libraries/chaucer/chaucer.cpp"
#include "ftoa.h"

const uint8_t flashnum PROGMEM = 0;

// const char hello[] PROGMEM = 
// const char hello[] PROGTEXT =
const char hello[] PROGDATA = 
  "Congratulations: your bootloader is pgm_write_page() enabled - trying a few tricks:\n";

const char resetprotected[] PROGDATA = 
  "Reset is write protected\n";

const char resetnotprotected[] PROGDATA = 
  "Vector bootloader without reset protection\n";

const char bootprotected[] PROGDATA = 
  "Bootloader is write protected\n";

#if SPM_PAGESIZE & (SPM_PAGESIZE-1)
#error SPM_PAGESIZE is not a power of 2; code will not work.
#endif

#define printtale(tale)  stdio_puts_Pnf(pgm_addr(tale[0]))

// Test whether flash region where is write protected
uint8_t testprotected(uint8_t *sram, progmem_t where) {
  uint8_t was[2], now[2];

  // Copy from flash, modify, write back, check and repair
  memcpy_Pnf(sram, where, SPM_PAGESIZE);
  memcpy(was, sram, sizeof was);
  if(where) {
    sram[0] = ~sram[0]; sram[1] = ~sram[1];
  } else { // Bootloader protects reset vector by writing correct rjmp
    sram[0] = 0; sram[1] = 0;
  }
  urbootPageWrite(sram, where);
  memcpy_Pnf(now, where, sizeof now);
  if(memcmp(was, now, sizeof was)) {
    memcpy(sram, was, sizeof was);
    urbootPageWrite(sram, where);
    if(where)
      return 0;
  }

  return where || now[0] || now[1];
}

// Not using printf_q() or similar as PROGMEM constants could be stored higher than 64 kB for large devices
void setup() {
  boardInit(1, "chaucer...\n");
  uint8_t ch;
  progmem_t pgm, base;

  // Does my own bootloader support pgm_write_page()?
  uint16_t capver = capver_ur();
  uint8_t ver = verbyte_cv(capver);
  if((ver > 070 && ver <= 076 && ispgmwritepage076_cv(capver)) || (ver >= 077 && rjmpwp_ur() != ub_ret_opcode)) {
    uint8_t sram[SPM_PAGESIZE];
    char space[10];

    printtale(hello);
    if(isvectorbl_cv(capver)) {
      if(testprotected(sram, 0))
        printtale(resetprotected);
      else
        printtale(resetnotprotected);
    }
      
    if(ver >= 075 && testprotected(sram, (progmem_t) (urstoreBelowBoot()+1)))
      printtale(bootprotected);
  
    // change the flashnum PROGMEM variable by reprogramming the application
    pgm = pgm_addr(flashnum);
    ch = pgm_read_b1(pgm);
    stdio_puts_qnf("old flashnum: ");
    utoa(space, (uint16_t) ch); stdio_puts(space);
    stdio_putc('\n');
    // start of progmem page
    base = (progmem_t) ((uintpgm_t) pgm & ~(uintpgm_t)(SPM_PAGESIZE-1));
    memcpy_Pnf(sram, base, SPM_PAGESIZE);
    // change copy of variable flashnum in sram
    sram[(spm_pagesize_store_t) (uintpgm_t) pgm & (SPM_PAGESIZE-1)]++;
    // overwrite flash with modified entry
    urbootPageWrite(sram, base);
    ch = pgm_read_b1(pgm);
    stdio_puts_qnf("new flashnum: ");
    utoa(space, (uint16_t) ch); stdio_puts(space);
    stdio_putc('\n');

    // this is fun: change the first full mem page in the first Knight's tale string
    pgm = pgm_addr(knightsTale0) + SPM_PAGESIZE;
    // start of progmem page
    base = (progmem_t) ((uintpgm_t) pgm & ~(uintpgm_t)(SPM_PAGESIZE-1));
    memcpy_Pnf(sram, base, SPM_PAGESIZE);
    if(ch == 0)
      ch=1;
    // change case of every ch-th character
    for(int16_t i=0; i<SPM_PAGESIZE; i+=ch)
      if(sram[i] > 0x40)
        sram[i] ^= 0x20;
    urbootPageWrite(sram, base);

    progmem_t end = base+SPM_PAGESIZE;

    for(pgm = pgm_addr(knightsTale0); (ch=pgm_read_b1(pgm)); pgm++) {
      if(pgm >= base && pgm < end)
       _delay_approx_ms(20);
      if(pgm == end) {
        stdio_puts_qnf("\n\nOdd cases in the preceding ");
        utoa(space, (uint16_t) SPM_PAGESIZE); stdio_puts(space);
        stdio_puts_qnf(" chars are a pgm_write_page() demo. Press a key.\n");
        stdio_getc();
      }
      stdio_putc(ch);
    }
  } else
    printtale(knightsTale0);

#if FLASHEND >= 65535U
  printtale(knightsTale1);
#endif
#if FLASHEND >= 131071UL
  printtale(knightsTale2);
  printtale(knightsTale3);
#endif
#if FLASHEND >= 262143UL
  printtale(millersTale0);
  printtale(millersTale1);
  printtale(wifeOfBathsTale0);
  printtale(wifeOfBathsTale1);
  printtale(wifeOfBathsTale2);
  printtale(cooksTale);
#endif

 stdio_puts_qnf("End of urboot demo\n");
}


void loop() {
}
